<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 03/25/2004 *
 * ===============                              Last change: 09/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : extensions.php                                   *
 * -------------------------------------------------------------------- *
 * Short description : Extension management                             *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Erweiterungen-Management                         *
 * -------------------------------------------------------------------- *
 * $Revision:: 856                                                    $ *
 * $Date:: 2009-03-06 20:24:32 +0100 (Fr, 06. Mär 2009)              $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author:: stelzi                                                   $ *
 * Needs to be in all Files and every File needs "svn propset           *
 * svn:keywords Date Revision" (autoprobset!) at least!!!!!!            *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2008 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	$INC = substr(dirname(__FILE__), 0, strpos(dirname(__FILE__), "/inc") + 4) . "/security.php";
	require($INC);
}

// Load the extension and maybe found language and function files.
function LOAD_EXTENSION ($ext_name, $EXT_LOAD_MODE = "", $EXT_VER = "", $dry_run = false) {
	global $INC_POOL;

	// Init array
	$INC_POOL = array();

	// Is the extension already loaded?
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "Loading extension {$ext_name}, mode={$EXT_LOAD_MODE}, ver={$EXT_VER}.");
	if ((isset($GLOBALS['ext_loaded']['ext'][$ext_name])) && (empty($EXT_LOAD_MODE))) {
		// Debug message
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Extension %s already loaded.", $ext_name));
		return false;
	} // END - if

	// Construct include filename and FQFN for extension file
	$INC = sprintf("inc/extensions/ext-%s.php", $ext_name);
	$FQFN = constant('PATH') . $INC;

	// Is the extension file NOT there?
	if (!INCLUDE_READABLE($INC)) {
		// Debug message
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Extension %s not found.", $ext_name));

		// Abort here
		return false;
	} // END - if

	// Construct FQFN for language file
	$langInclude = sprintf("inc/language/%s_%s.php", $ext_name, GET_LANGUAGE());

	// Is this include there?
	if ((FILE_READABLE($langInclude)) && (!isset($GLOBALS['ext_loaded']['lang'][$ext_name]))) {
		// Then load it
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "--- Language loaded.");
		$GLOBALS['ext_loaded']['lang'][$ext_name] = true;
		LOAD_INC_ONCE($langInclude);
	} // END - if

	// Construct FQFN for functions file
	$funcsInclude = sprintf("inc/libs/%s_functions.php", $ext_name);

	// Is this include there?
	if ((FILE_READABLE($funcsInclude)) && (!isset($GLOBALS['ext_loaded']['funcs'][$ext_name]))) {
		// Then load it
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "--- Functions loaded.");
		$GLOBALS['ext_loaded']['funcs'][$ext_name] = true;
		LOAD_INC_ONCE($funcsInclude);
	} // END - if

	// Extensions are not deprecated by default
	EXT_SET_DEPRECATED("N");

	// Extensions are not always active by default
	EXT_SET_ALWAYS_ACTIVE("N");

	// By default an extension update does not depend on other extensions
	EXT_SET_UPDATE_DEPENDS("");

	// Extension update notes
	EXT_SET_UPDATE_NOTES("");

	// Include the extension file
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "Extension loaded.");
	require($FQFN);

	// Is this extension deprecated?
	if (EXT_GET_DEPRECATED() == "Y") {
		// Deactivate the extension
		DEACTIVATE_EXTENSION($ext_name);

		// Abort here
		return false;
	} // END - if

	// Mark it as loaded in normal mode
	if (empty($EXT_LOAD_MODE)) {
		// Mark it now...
		$GLOBALS['ext_loaded']['ext'][$ext_name] = true;
	} // END - if

	// All fine!
	return true;
}

// Registeres an extension and possible update depencies
function EXTENSION_REGISTER ($ext_name, $task_id, $dry_run = false, $logout = true) {
	global $INC_POOL;
	global $NOTES;

	// This shall never do a non-admin user!
	if (!IS_ADMIN()) return false;

	// Is this extension already installed?
	if (EXT_IS_ACTIVE($ext_name)) return false;

	// Init queries
	INIT_SQLS();

	// Init variables
	$ret = false;
	$NOTES = "";
	$INC_POOL = array();

	// By default we have no failures
	EXT_SET_REPORTS_FAILURE(false);

	// Does this extension exists?
	if (LOAD_EXTENSION($ext_name, "register", "", $dry_run)) {
		// And run possible updates
		$history = EXT_GET_VER_HISTORY();
		foreach ($history as $ver) {
			//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "ext_name={$ext_name}, ext_ver={$ver}");
			// Load extension in update mode
			LOAD_EXTENSION($ext_name, "update", $ver, $dry_run);

			// Add update notes to our output
			$NOTES .= ADD_EXTENSION_NOTES($ver);
		} // END - foreach

		// Does this extension depends on an outstanding update of another update?
		if (EXT_GET_UPDATE_DEPENDS() != "") {
			// Backup SQL commands and clear current
			$SQLs2 = GET_SQLS();
			INIT_SQLS();
			$test  = false;
			$ext_update = EXT_GET_UPDATE_DEPENDS();

			// Check for required file
			if (LOAD_EXTENSION($ext_update, "register", "", $dry_run)) {
				// If versions mismatch update extension first
				$ext_ver = GET_EXT_VERSION($ext_update);

				// Extension version set? If empty the extension is not registered
				if (empty($ext_ver)) {
					// Extension not registered so far so first load task's ID...
					$task = DETERMINE_EXTENSION_TASK_ID($ext_update);

					// Entry found?
					if ($task > 0) {
						// Try to register the extension
						$test = EXTENSION_REGISTER($ext_update, $task, $dry_run, false);
					} // END - if
				} elseif ($ext_ver != EXT_GET_VERSION()) {
					// Ok, update this extension now
					EXTENSION_UPDATE($ext_update, $ext_ver, $dry_run);

					// All okay!
					$test = true;
				} else {
					// Nothing to register / update before...
					$test = true;
				}
			} else {
				// Required file for update does not exists!
				$test = true;
				// But this is fine for the first time...
			}

			// Finally restore previous SQLs
			SET_SQLS($SQLs2); unset($SQLs2);
		} else {
			// Does not depend on an other extension
			$test = true;
		}

		// Switch back to register mode
		$EXT_LOAD_MODE = "register";

		// Remains true if extension registration reports no failures
		$test = (($test === true) && (EXT_GET_REPORTS_FAILURE() === false));

		// Does everthing before wents ok?
		if ($test) {
			// "Dry-run-mode" activated?
			if (!$dry_run) {
				// Run installation pre-installation filters
				runFilterChain('pre_extension_installed', array('dry_run' => $dry_run));

				// Register extension
				SQL_QUERY_ESC("INSERT INTO `{!_MYSQL_PREFIX!}_extensions` (ext_name, ext_active, ext_version) VALUES ('%s','%s','%s')",
					array($ext_name, EXT_GET_ALWAYS_ACTIVE(), EXT_GET_VERSION()), __FILE__, __LINE__);

				// Remove cache file(s) if extension is active
				runFilterChain('post_extension_installed', array('ext_name' => $ext_name, 'task_id' => $task_id, 'inc_pool' => $INC_POOL));

				// In normal mode return a true on success
				$ret = true;
			} else {
				// Rewrite SQL command to keep { and } inside
				foreach (GET_SQLS() as $key => $sql) {
					$sql = str_replace('{', "&#123;", str_replace('}', "&#125;", $sql));
					SET_SQL_KEY($key, $sql);
				} // END - foreach

				// In  "dry-run" mode return array with all SQL commands
				$ret = GET_SQLS();

				// Remove all SQL commands
				UNSET_SQLS();
			}
		} else {
			// No, an error occurs while registering extension :-(
			$ret = false;
		}
	} elseif (($task_id > 0) && (!empty($ext_name))) {
		// Remove task from system when id and extension's name is valid
		SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{!_MYSQL_PREFIX!}_task_system` WHERE id=%s AND `status`='NEW' LIMIT 1",
			array(bigintval($task_id)), __FILE__, __LINE__);
	}

	// Is this the sql_patches?
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ":{$ext_name}/{$EXT_LOAD_MODE}");
	if (($ext_name == "sql_patches") && (($EXT_LOAD_MODE == "register") || ($EXT_LOAD_MODE == "remove")) && (!$dry_run) && ($test)) {
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": LOAD!");
		if ($logout === true) {
			// Then redirect to logout
			LOAD_URL("modules.php?module=admin&amp;logout=1&amp;".$EXT_LOAD_MODE."=sql_patches");
		} else {
			// Add temporary filter
			REGISTER_FILTER('shutdown', 'REDIRECT_TO_LOGOUT_SQL_PATCHES', true, true);
			$GLOBALS['ext_load_mode'] = $EXT_LOAD_MODE;
		}
	} // END - if

	// Return status code
	return $ret;
}

// Run SQL queries for given extension id
// @TODO Change from ext_id to ext_name (not just even the variable! ;-) )
function EXTENSION_RUN_SQLS ($ext_id, $load_mode) {
	// This shall never do a non-admin user!
	if (!IS_ADMIN()) return false;

	// Init array
	INIT_SQLS();

	// By default no SQL has been executed
	$sqlRan = false;

	// Get extension's name
	$ext_name = GET_EXT_NAME($ext_id);
	// If it is not set then maybe there is no extension for that ID number
	if (empty($ext_name)) return false;

	// Load extension in detected mode
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ":ext_name[{$ext_id}]={$ext_name}");
	LOAD_EXTENSION($ext_name, $load_mode, "", false);

	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ":SQLs::count=".COUNT_SQLS()."");
	if ((IS_SQLS_VALID() && (COUNT_SQLS() > 0))) {
		// Run SQL commands...
		runFilterChain('run_sqls');

		// Removal mode?
		if ($load_mode == "remove") {
			// Delete this extension (remember to remove it from your server *before* you click on welcome!
			SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{!_MYSQL_PREFIX!}_extensions` WHERE ext_name='%s' LIMIT 1",
				array($ext_name), __FILE__, __LINE__);
		} // END - if
	} // END - if

	// Remove cache file(s) if extension is active
	if (((EXT_IS_ACTIVE("cache")) || (GET_EXT_VERSION("cache") != "")) && (((SQL_AFFECTEDROWS() == 1)) || ($sqlRan === true) || ($load_mode == "activate") || ($load_mode == "deactivate"))) {
		// Run filters
		runFilterChain('post_extension_run_sql', $ext_name);
	} // END - if

	// Is this the sql_patches?
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": {$ext_id}/{$ext_name}/{$load_mode}");
	if (($ext_name == "sql_patches") && (($load_mode == "register") || ($load_mode == "remove"))) {
		// Then redirect to logout
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": LOAD!");
		LOAD_URL("modules.php?module=admin&amp;logout=1&amp;".$load_mode."=sql_patches");
	} // END - if
}

// Check if given extension is active
function EXT_IS_ACTIVE ($ext_name) {
	// Extensions are all inactive during installation
	if ((!isInstalled()) || (isInstalling()) || (empty($ext_name))) return false;

	// Not active is the default
	$active = "N";

	// Check cache
	if (isset($GLOBALS['cache_array']['extensions']['ext_active'][$ext_name])) {
		// Load from cache
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "CACHE! ext_name={$ext_name}");
		$active = $GLOBALS['cache_array']['extensions']['ext_active'][$ext_name];

		// Count cache hits
		incrementConfigEntry('cache_hits');
	} elseif (isset($GLOBALS['ext_loaded'][$ext_name])) {
		// Extension is loaded!
		die("LOADED:$ext_name");
	} elseif (($ext_name == "cache") || (GET_EXT_VERSION("cache") == "")) {
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "DB! ext_name={$ext_name}");
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_active FROM `{!_MYSQL_PREFIX!}_extensions` WHERE ext_name='%s' LIMIT 1",
			array($ext_name), __FILE__, __LINE__);

		// Entry found?
		if (SQL_NUMROWS($result) == 1) {
			// Load entry
			list($active) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);

		// Write cache array
		//* DEBUG: */ echo $ext_name."[DB]: {$active}");
		$GLOBALS['cache_array']['extensions']['ext_active'][$ext_name] = $active;
	} else {
		// Extension not active!
		//* DEBUG: */ echo $ext_name.": Not active!");
		$GLOBALS['cache_array']['extensions']['ext_active'][$ext_name] = "N";
	}

	// Debug message
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, " ext_name={$ext_name},active={$active}");

	// Is this extension activated? (For admins we always have active extensions...)
	return ($active == "Y");
}
// Get version from extensions
function GET_EXT_VERSION ($ext_name) {
	// By default no extension is found
	$ext_ver = false;

	// Extensions are all inactive during installation
	if ((!isInstalled()) || (isInstalling())) return "";
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": ext_name={$ext_name}");

	// Is the cache written?
	if (isset($GLOBALS['cache_array']['extensions']['ext_version'][$ext_name])) {
		// Load data from cache
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": CACHE!");
		$ext_ver = $GLOBALS['cache_array']['extensions']['ext_version'][$ext_name];

		// Count cache hits
		incrementConfigEntry('cache_hits');
	} elseif (!isCacheInstanceValid()) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_version FROM `{!_MYSQL_PREFIX!}_extensions` WHERE ext_name='%s' LIMIT 1",
			array($ext_name), __FILE__, __LINE__);
		//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": DB - ".SQL_NUMROWS($result)."");

		// Is the extension there?
		if (SQL_NUMROWS($result) == 1) {
			// Load entry
			list($ext_ver) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);

		// Set cache
		$GLOBALS['cache_array']['extensions']['ext_version'][$ext_name] = $ext_ver;
	}

	// Return result
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, ": ret={$ext_ver}");
	return $ext_ver;
}

// Updates a given extension with current extension version to latest version
function EXTENSION_UPDATE ($ext_name, $ext_ver, $dry_run = false) {
	// This shall never do a non-admin user!
	global $NOTES;
	global $INC_POOL;

	// Init arrays
	INIT_SQLS();
	$INC_POOL = array();

	// Init notes
	$NOTES = "";

	// Only admins are allowed to update extensions
	if ((!IS_ADMIN()) || (empty($ext_name))) return false;

	// Load extension in test mode
	LOAD_EXTENSION($ext_name, "test", $ext_ver, $dry_run);

	// Save version history
	$history = EXT_GET_VER_HISTORY();

	// Remove old SQLs array to prevent possible bugs
	INIT_SQLS();

	// Check if version is updated
	if (((EXT_GET_VERSION() != $ext_ver) || ($dry_run)) && (is_array($history))) {
		// Search for starting point
		$start = array_search($ext_ver, $history);

		// And load SQL queries in order of version history
		for ($idx = ($start + 1); $idx < count($history); $idx++) {
			// Set extension version
			$GLOBALS['cache_array']['update_ver'][$ext_name] = $history[$idx];

			// Load again...
			LOAD_EXTENSION($ext_name, "update", $GLOBALS['cache_array']['update_ver'][$ext_name], $dry_run);

			if (EXT_GET_UPDATE_DEPENDS() != "") {
				// Backup current SQL queries
				$GLOBALS['cache_array']['update_sqls'][$ext_name] = GET_SQLS();

				// Is the extension there?
				if (GET_EXT_VERSION(EXT_GET_UPDATE_DEPENDS()) != "") {
					// Update another extension first!
					$test = EXTENSION_UPDATE(EXT_GET_UPDATE_DEPENDS(), GET_EXT_VERSION(EXT_GET_UPDATE_DEPENDS()), $dry_run);
				} else {
					// Register new extension
					$test = EXTENSION_REGISTER(EXT_GET_UPDATE_DEPENDS(), 0, $dry_run, false);
				}

				// Restore previous SQL queries
				SET_SQLS($GLOBALS['cache_array']['update_sqls'][$ext_name]);
				unset($GLOBALS['cache_array']['update_sqls'][$ext_name]);
			} // END - if

			// Add notes
			$NOTES .= ADD_EXTENSION_NOTES($GLOBALS['cache_array']['update_ver'][$ext_name]);
		} // END - for

		// In real-mode execute any existing includes
		if (!$dry_run) {
			$GLOBALS['cache_array']['inc_pool'][$ext_name] = $INC_POOL;
			runFilterChain('load_includes', $INC_POOL);
			$INC_POOL = $GLOBALS['cache_array']['inc_pool'][$ext_name];
			unset($GLOBALS['cache_array']['inc_pool'][$ext_name]);
		} // END - if

		// Run SQLs
		runFilterChain('run_sqls', array('dry_run' => $dry_run));

		if (!$dry_run) {
			// Create task
			CREATE_EXTENSION_UPDATE_TASK(GET_CURRENT_ADMIN_ID(), $ext_name, $GLOBALS['cache_array']['update_ver'][$ext_name], SQL_ESCAPE($NOTES));

			// Update extension's version
			SQL_QUERY_ESC("UPDATE `{!_MYSQL_PREFIX!}_extensions` SET ext_version='%s' WHERE ext_name='%s' LIMIT 1",
				array($GLOBALS['cache_array']['update_ver'][$ext_name], $ext_name), __FILE__, __LINE__);

			// Remove arrays
			UNSET_SQLS();
			unset($GLOBALS['cache_array']['update_ver'][$ext_name]);

			// Run filters on success extension update
			runFilterChain('extension_update', $ext_name);
		} // END - if
	} // END - if
}

// Output verbose SQL table for extension
function EXTENSION_VERBOSE_TABLE ($queries = array(), $title = "", $dashed = "", $switch = false, $width = "100%") {
	// Empty title?
	if (empty($title)) {
		// Then fix it to default
		$title = getMessage('ADMIN_SQLS_EXECUTED_ON_REMOVAL');
	} // END - if

	// Are there some queries in $queries?
	if (count($queries) > 0) {
		// Then use them instead!
		SET_SQLS($queries);
	} // END - if

	// Init variables
	$SW = 2; $i = 1;
	$OUT = "";

	// Do we have queries?
	if ((IS_SQLS_VALID()) && (GET_EXT_VERSION("sql_patches") >= "0.0.7") && (getConfig('verbose_sql') == "Y")) {
		foreach (GET_SQLS() as $idx => $sql) {
			// Trim out spaces
			$sql = trim($sql);

			// Output command if set
			if (!empty($sql)) {
				// Prepare output for template
				$content = array(
					'sw'  => $SW,
					'i'   => $i,
					'sql' => $sql
				);

				// Load row template
				$OUT .= LOAD_TEMPLATE("admin_ext_sql_row", true, $content);

				// Switch color and count up
				$SW = 3 - $SW;
				$i++;
			} // END - if
		} // END - foreach

		// Prepare content for template
		$content = array(
			'width'  => $width,
			'dashed' => $dashed,
			'title'  => $title,
			'out'    => $OUT
		);

		// Load main template
		$OUT = LOAD_TEMPLATE("admin_ext_sql_table", true, $content);
	} elseif ((GET_EXT_VERSION("sql_patches") >= "0.0.7") && (getConfig('verbose_sql') == "Y")) {
		// No addional SQL commands to run
		$OUT = LOAD_TEMPLATE("admin_settings_saved", true, getMessage('ADMIN_NO_ADDITIONAL_SQLS'));
	} // END - if

	// Return output
	return $OUT;
}

// Get extension name from id
function GET_EXT_NAME ($ext_id) {
	// Init extension name
	$ret = "";

	// Is cache there?
	if (isset($GLOBALS['cache_array']['extensions']['ext_name'][$ext_id])) {
		// Load from cache
		$ret = $GLOBALS['cache_array']['extensions']['ext_name'][$ext_id];

		// Count cache hits
		incrementConfigEntry('cache_hits');
	} elseif (!EXT_IS_ACTIVE("cache")) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_name FROM `{!_MYSQL_PREFIX!}_extensions` WHERE id=%s LIMIT 1",
			array(bigintval($ext_id)), __FILE__, __LINE__);
		list($ret) = SQL_FETCHROW($result);
		SQL_FREERESULT($result);
	}
	return $ret;
}

// Get extension id from name
function GET_EXT_ID ($ext_name) {
	// Init ID number
	$ret = 0;
	if (isset($GLOBALS['cache_array']['extensions']['ext_id'][$ext_name])) {
		// Load from cache
		$ret = $GLOBALS['cache_array']['extensions']['ext_id'][$ext_name];

		// Count cache hits
		incrementConfigEntry('cache_hits');
	} elseif (!EXT_IS_ACTIVE("cache")) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT id FROM `{!_MYSQL_PREFIX!}_extensions` WHERE ext_name='%s' LIMIT 1",
			array($ext_name), __FILE__, __LINE__);
		list($ret) = SQL_FETCHROW($result);
		SQL_FREERESULT($result);
	}

	// Return value
	return $ret;
}

// Activate given extension
function ACTIVATE_EXTENSION ($ext_name) {
	// Activate the extension
	SQL_QUERY_ESC("UPDATE `{!_MYSQL_PREFIX!}_extensions` SET ext_active='Y' WHERE ext_name='%s' LIMIT 1",
		array($ext_name), __FILE__, __LINE__);

	// Extension has been activated?
	if (SQL_AFFECTEDROWS() == 1) {
		// Then run all queries
		EXTENSION_RUN_SQLS(GET_EXT_ID($ext_name), "activate");
	} // END - if
}

// Deactivate given extension
function DEACTIVATE_EXTENSION($ext_name) {
	// Activate the extension
	SQL_QUERY_ESC("UPDATE `{!_MYSQL_PREFIX!}_extensions` SET ext_active='N' WHERE ext_name='%s' LIMIT 1",
		array($ext_name), __FILE__, __LINE__);

	// Extension has been activated?
	if (SQL_AFFECTEDROWS() == 1) {
		// Then run all queries
		EXTENSION_RUN_SQLS(GET_EXT_ID($ext_name), "deactivate");

		// Create new task
		CREATE_EXTENSION_DEACTIVATION_TASK($ext_name);

		// Notify the admin
		SEND_ADMIN_NOTIFICATION(
			getMessage('ADMIN_SUBJECT_EXTENSION_DEACTIVATED'),
			"admin_ext_deactivated",
			array('ext_name' => $ext_name)
		);
	} // END - if
}

// Checks wether the extension is older than given
function EXT_VERSION_IS_OLDER ($ext_name, $ext_ver) {
	// Get current extension version
	$currVersion = GET_EXT_VERSION($ext_name);

	// Remove all dots from both versions
	$currVersion = str_replace(".", "", $currVersion);
	$ext_ver = str_replace(".", "", $ext_ver);

	// Now compare both and return the result
	return ($currVersion < $ext_ver);
}

// Creates a new task for updated extension
function CREATE_EXTENSION_UPDATE_TASK ($admin_id, $ext_name, $ext_ver, $notes) {
	// Create subject line
	$subject = "[UPDATE-".$ext_name."-".$ext_ver.":] {--ADMIN_UPDATE_EXT_SUBJ--}";

	// Is the extension there?
	if (GET_EXT_VERSION($ext_name) != "") {
		// Check if task is not there
		if (DETERMINE_TASK_ID_BY_SUBJECT($subject) == 0) {
			// Task not created so it's a brand-new extension which we need to register and create a task for!
			SQL_QUERY_ESC("INSERT INTO `{!_MYSQL_PREFIX!}_task_system` (assigned_admin, userid, status, task_type, subject, text, task_created) VALUES ('%s','0','NEW','EXTENSION_UPDATE','%s','%s', UNIX_TIMESTAMP())",
				array($admin_id, $subject, $notes), __FILE__, __LINE__);
		} // END - if
	} // END - if
}

// Creates a new task for newly installed extension
function CREATE_NEW_EXTENSION_TASK ($admin_id, $subject, $ext) {
	// Not installed and do we have created a task for the admin?
	if ((DETERMINE_TASK_ID_BY_SUBJECT($subject) == 0) && (GET_EXT_VERSION($ext) == "")) {
		// Template file
		$tpl = sprintf("%stemplates/%s/html/ext/ext_%s.tpl",
			constant('PATH'),
			GET_LANGUAGE(),
			$ext
		);

		// Load text for task
		if (FILE_READABLE($tpl)) {
			// Load extension's own text template (HTML!)
			$msg = LOAD_TEMPLATE("ext_".$ext, true);
		} else {
			// Load default message
			$msg = LOAD_TEMPLATE("admin_new_ext", "", 0);
		}

		// Task not created so it's a brand-new extension which we need to register and create a task for!
		SQL_QUERY_ESC("INSERT INTO `{!_MYSQL_PREFIX!}_task_system` (assigned_admin, userid, status, task_type, subject, text, task_created)
VALUES (%s,0,'NEW','EXTENSION','%s','%s',UNIX_TIMESTAMP())",
			array(
				$admin_id,
				$subject,
				SQL_ESCAPE($msg),
			),  __FILE__, __LINE__, true, false
		);
	} // END - if
}

// Creates a task for automatically deactivated (deprecated) extension
function CREATE_EXTENSION_DEACTIVATION_TASK ($ext) {
	// Create subject line
	$subject = sprintf("[%s:] %s", $ext, getMessage('TASK_SUBJ_EXTENSION_DEACTIVATED'));

	// Not installed and do we have created a task for the admin?
	if ((DETERMINE_TASK_ID_BY_SUBJECT($subject) == 0) && (GET_EXT_VERSION($ext) != "")) {
		// Task not created so add it
		SQL_QUERY_ESC("INSERT INTO `{!_MYSQL_PREFIX!}_task_system` (assigned_admin, userid, status, task_type, subject, text, task_created)
VALUES (0,0,'NEW','EXTENSION_DEACTIVATION','%s','%s',UNIX_TIMESTAMP())",
			array(
				$subject,
				SQL_ESCAPE(LOAD_TEMPLATE("task_ext_deactivated", true, $ext)),
			),  __FILE__, __LINE__, true, false
		);
	} // END - if
}

// Checks if the module has a menu
function MODULE_HAS_MENU ($mod, $forceDb = false) {
	// All is false by default
	$ret = false;

	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):mod={$mod},cache=".GET_EXT_VERSION("cache")."<br />\n";
	if (GET_EXT_VERSION("cache") >= "0.1.2") {
		// Cache version is okay, so let's check the cache!
		if (isset($GLOBALS['cache_array']['modules']['has_menu'][$mod])) {
			// Check module cache and count hit
			$ret = ($GLOBALS['cache_array']['modules']['has_menu'][$mod] == "Y");
			incrementConfigEntry('cache_hits');
		} elseif (isset($GLOBALS['cache_array']['extensions']['ext_menu'][$mod])) {
			// Check cache and count hit
			$ret = ($GLOBALS['cache_array']['extensions']['ext_menu'][$mod] == "Y");
			incrementConfigEntry('cache_hits');
		} elseif ((IS_ADMIN()) && ($mod == "admin")) {
			// Admin module has always a menu!
			$ret = true;
		}
	} elseif ((GET_EXT_VERSION("sql_patches") >= "0.3.6") && ((!EXT_IS_ACTIVE("cache")) || ($forceDb === true))) {
		// Check database for entry
		$result = SQL_QUERY_ESC("SELECT has_menu FROM `{!_MYSQL_PREFIX!}_mod_reg` WHERE module='%s' LIMIT 1",
			array($mod), __FILE__, __LINE__);

		// Entry found?
		if (SQL_NUMROWS($result) == 1) {
			// Load "has_menu" column
			list($has_menu) = SQL_FETCHROW($result);

			// Fake cache... ;-)
			$GLOBALS['cache_array']['extensions']['ext_menu'][$mod] = $has_menu;

			// Does it have a menu?
			$ret = ($has_menu == "Y");
		} // END  - if

		// Free memory
		SQL_FREERESULT($result);
	} elseif (GET_EXT_VERSION("sql_patches") == "") {
		// No sql_patches installed, so maybe in admin area or no admin registered?
		$ret = (((IS_ADMIN()) || (!isAdminRegistered())) && ($mod == "admin")); // Then there is a menu!
	}

	// Return status
	//* DEBUG: */ var_dump($ret);
	return $ret;
}

// Determines the task id for given extension
function DETERMINE_EXTENSION_TASK_ID ($ext_name) {
	// Default is not found
	$task_id = 0;

	// Search for extension task's id
	$result = SQL_QUERY_ESC("SELECT id FROM `{!_MYSQL_PREFIX!}_task_system` WHERE task_type='EXTENSION' AND subject='[%s:]' LIMIT 1",
		array($ext_name), __FILE__, __LINE__);

	// Entry found?
	if (SQL_NUMROWS($result) == 1) {
		// Task found so load task's ID and register extension...
		list($task_id) = SQL_FETCHROW($result);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return it
	return $task_id;
}

// Determines the task id for given subject
function DETERMINE_TASK_ID_BY_SUBJECT ($subject) {
	// Default is not found
	$task_id = 0;

	// Search for task id
	$result = SQL_QUERY_ESC("SELECT id FROM `{!_MYSQL_PREFIX!}_task_system` WHERE subject LIKE '%s%%' LIMIT 1",
		array($subject), __FILE__, __LINE__);

	// Entry found?
	if (SQL_NUMROWS($result) == 1) {
		// Task found so load task's ID and register extension...
		list($task_id) = SQL_FETCHROW($result);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return it
	return $task_id;
}

// Add updates notes for given version
function ADD_EXTENSION_NOTES ($ver) {
	// Init notes/content
	$out = ""; $content = array();

	// Is do we have verbose output enabled?
	if ((getConfig('verbose_sql') == "Y") || (!EXT_IS_ACTIVE("sql_patches"))) {

		// Update notes found?
		if (EXT_GET_UPDATE_NOTES() != "") {
			// Update notes found
			$content = array(
				'ver'   => $ver,
				'notes' => EXT_GET_UPDATE_NOTES()
			);
			EXT_SET_UPDATE_NOTES("");
		} elseif (($ver == "0.0") || ($ver == "0.0.0")) {
			// Initial release
			$content = array(
				'ver'   => $ver,
				'notes' => getMessage('INITIAL_RELEASE')
			);
		} else {
			// No update notes found!
			$content = array(
				'ver'   => $ver,
				'notes' => getMessage('NO_UPDATE_NOTES')
			);
		}

		// Load template
		$out = LOAD_TEMPLATE("admin_ext_notes", true, $content);
	} // END - if

	// Return the output
	return $out;
}

// Getter for CSS files array
function EXT_GET_CSS_FILES () {
	// By default no additional CSS files are found
	$cssFiles = array();

	// Is the array there?
	if (isset($GLOBALS['css_files'])) {
		// Then use it
		$cssFiles = $GLOBALS['css_files'];
	} // END - if

	// Return array
	return $cssFiles;
}

// Init CSS files array
function EXT_INIT_CSS_FILES () {
	// Simply init it
	$GLOBALS['css_files'] = array();
}

// Add new entry
function EXT_ADD_CSS_FILE ($file) {
	// Is the array there?
	if (!isset($GLOBALS['css_files'])) {
		// Then auto-init them
		EXT_INIT_CSS_FILES();
	} // END - if

	// Add the entry
	$GLOBALS['css_files'][] = $file;
}

// Setter for EXT_ALWAYS_ACTIVE flag
function EXT_SET_ALWAYS_ACTIVE ($active) {
	$GLOBALS['ext_always_active'] = $active;
}

// Getter for EXT_ALWAYS_ACTIVE flag
function EXT_GET_ALWAYS_ACTIVE () {
	return $GLOBALS['ext_always_active'];
}

// Setter for EXT_VERSION flag
function EXT_SET_VERSION ($version) {
	$GLOBALS['ext_version'] = $version;
}

// Getter for EXT_VERSION flag
function EXT_GET_VERSION () {
	return $GLOBALS['ext_version'];
}

// Setter for EXT_DEPRECATED flag
function EXT_SET_DEPRECATED ($deprecated) {
	$GLOBALS['ext_deprecated'] = $deprecated;
}

// Getter for EXT_DEPRECATED flag
function EXT_GET_DEPRECATED () {
	return $GLOBALS['ext_deprecated'];
}

// Setter for EXT_UPDATE_DEPENDS flag
function EXT_SET_UPDATE_DEPENDS ($updateDepends) {
	$GLOBALS['ext_update_depends'] = $updateDepends;
}

// Getter for EXT_UPDATE_DEPENDS flag
function EXT_GET_UPDATE_DEPENDS () {
	return $GLOBALS['ext_update_depends'];
}

// Setter for EXT_REPORTS_FAILURE flag
function EXT_SET_REPORTS_FAILURE ($reportsFailure) {
	$GLOBALS['ext_reports_failure'] = $reportsFailure;
}

// Getter for EXT_REPORTS_FAILURE flag
function EXT_GET_REPORTS_FAILURE () {
	return $GLOBALS['ext_reports_failure'];
}

// Setter for EXT_VER_HISTORY flag
function EXT_SET_VER_HISTORY ($verHistory) {
	$GLOBALS['ext_ver_history'] = (array) $verHistory;
}

// Getter for EXT_VER_HISTORY flag
function EXT_GET_VER_HISTORY () {
	return $GLOBALS['ext_ver_history'];
}

// Setter for EXT_UPDATE_NOTES flag
function EXT_SET_UPDATE_NOTES ($updateNotes) {
	$GLOBALS['ext_update_notes'] = (array) $updateNotes;
}

// Getter for EXT_UPDATE_NOTES flag
function EXT_GET_UPDATE_NOTES () {
	return $GLOBALS['ext_update_notes'];
}

//
?>
